# coding: utf-8

"""
MIT License

Copyright (c) 2018 Claude SIMON (https://q37.info/s/rmnmqd49)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import XDHqFaaS,XDHqSHRD,XDHqXML
from XDHqFaaS import launch, set_supplier, get_app_url, setBye, l

import os,sys
from collections import OrderedDict

if sys.version_info[0] == 2:
	import urllib
	def _encode(string):
		return urllib.quote(string)
elif sys.version_info[0] == 3:
	import urllib.parse
	def _encode(string):
		return urllib.parse.quote(string)
else:
	print("Unhandled python version!")
	os._exit(1)

_VOID = XDHqSHRD.RT_VOID
_STRING = XDHqSHRD.RT_STRING
_STRINGS = XDHqSHRD.RT_STRINGS
XML = XDHqXML.XML


def _split(keysAndValues):
	keys = []
	values = []

	for key in keysAndValues:
		keys.append(str(key))
		values.append(str(keysAndValues[key]))

	return [keys,values]

def _unsplit(keys,values):
	i = 0
	# With 'OrderedDict', the order of the items is keeped under Python 2.
	# This facilitates the retrieving of values by using 'values()' method.
	# Dicts are ordered by default under Python 3.
	keysAndValues = OrderedDict()
	length = len(keys)

	while i < length:
		keysAndValues[keys[i]] = values[i]
		i += 1

	return keysAndValues

def _getAssetPath():
	if not XDHqSHRD.isDev():
		throw("Should only be called in DEV context!!!")

	return os.path.join(os.path.realpath(os.path.join(os.environ["Q37_EPEIOS"].replace('h:','/cygdrive/h'),"tools/xdhq/examples/common/")),os.path.relpath(os.getcwd(),os.path.realpath(os.path.join(os.environ["Q37_EPEIOS"].replace('h:','/cygdrive/h'),"tools/xdhq/examples/PYH/"))))


def get_asset_filename(path):
	return os.path.join(_getAssetPath(),path )


broadcastAction = XDHqFaaS.broadcastAction


def _readXSLAsset(path):
	if (path.lstrip()[0]=='<'):
		return path.lstrip()
	else:
		return open(get_asset_filename(path) if XDHqSHRD.isDev() else path).read()

class DOM:
	def __init__(self,instance):
		self._dom = XDHqFaaS.DOM_FaaS(instance)

	def get_action(self):
		return self._dom.getAction()

	getAction = get_action

	def isQuitting(self):
		return self._dom.isQuitting()

	def _execute(self,type,script):
		return self._dom.call("Execute_1",type,script)

	# Last statement of 'script' MUST be 'undefined', or the thread will be killed.
	def executeVoid(self,script):
		return self._execute(_VOID,script)

	execute_void = executeVoid

	def executeString(self,script):
		return self._execute(_STRING,script)

	execute_string = executeString

	def executeStrings(self,script):
		return self._execute(_STRINGS,script)

	execute_strings = executeStrings

	def _raw_send(self,type,data):
		return self._dom.call("RawSend_1",type,data)

	# Last statement of 'data' MUST be 'undefined', or the thread will be killed.
	def rawSendVoid(self,data):
		return self._raw_send(_VOID,data)

	raw_send_void = rawSendVoid

	def rawSendString(self,data):
		return self._raw_send(_STRING,data)

	raw_send_string = rawSendString

	def rawSendStrings(self,data):
		return self._raw_send(_STRINGS,data)

	raw_send_strings = rawSendStrings	

	def flush(self):	# Returns when all the pending commands were executed.
		self._dom.call("Flush_1",_STRING)

	def alert(self,message):
		self._dom.call( "Alert_1",_STRING,str(message) )
		# For the return value being 'STRING' instead of 'VOID',
		# see the 'alert' primitive in 'XDHqXDH'.

	def confirm(self,message):
		return self._dom.call( "Confirm_1",_STRING,message ) == "true"

	def _handleLayout(self,variant,id,xml,xsl):
		#	If 'xslFilename' is empty, 'xml' contents HTML.
		# If 'xml' is HTML and uses the compressed form, if it has a root tag, only the children will be used.
		# The corresponding primitive returns a value, which is only used internally, hence the lack of 'return'.
		# It also serves to do some synchronisation.
		self._dom.call("HandleLayout_1",_STRING,variant,id,xml if isinstance(xml,str) else xml.toString(),xsl)

	def prependLayout(self,id,html):	# Deprecated!
		self._handleLayout("Prepend",id,html,"")

	prepend_layout = prependLayout	# Deprecated!

	def setLayout(self,id,html):	# Deprecated!
		self._handleLayout("Set",id,html,"")

	set_layout = setLayout	# Deprecated!

	def appendLayout(self,id,html):	# Deprecated!
		self._handleLayout("Append",id,html,"")

	append_layout = appendLayout	# Deprecated!

	def _handleLayoutXSL(self,variant,id,xml,xsl):	# Deprecated!
		xslURL = xsl

		if True:	# Testing if 'SlfH' or 'FaaS' mode when available.
			xslURL = "data:text/xml;charset=utf-8," + _encode(_readXSLAsset(xsl))

		self._handleLayout(variant,id,xml,xslURL )

	def prependLayoutXSL(self,id,xml,xsl):	# Deprecated!
		self._handleLayoutXSL("Prepend",id,xml,xsl)

	prepend_layout_XSL = prependLayoutXSL	# Deprecated!

	def setLayoutXSL(self,id,xml,xsl):	# Deprecated!
		self._handleLayoutXSL("Set",id,xml,xsl)

	set_layout_XSL = setLayoutXSL	# Deprecated!

	def appendLayoutXSL(self,id,xml,xsl):	# Deprecated!
		self._handleLayoutXSL("Append",id,xml,xsl)

	append_layout_XSL = appendLayoutXSL	# Deprecated!

	def _layout(self,variant,id,xml,xsl):
		if xsl:
			xsl = "data:text/xml;charset=utf-8," + _encode(_readXSLAsset(xsl))

		self._dom.call("HandleLayout_1",_STRING,variant,id,xml if isinstance(xml,str) else xml.toString(),xsl)

	def before(self,id,xml,xsl=""):
		self._layout("beforebegin",id,xml,xsl)

	def begin(self,id,xml,xsl=""):
		self._layout("afterbegin",id,xml,xsl)

	def inner(self,id,xml,xsl=""):
		self._layout("inner",id,xml,xsl)

	def end(self,id,xml,xsl=""):
		self._layout("beforeend",id,xml,xsl)

	def after(self,id,xml,xsl=""):
		self._layout("afterend",id,xml,xsl)

	# Deprecated
	def getContents(self,ids):
		return _unsplit(ids,self._dom.call("GetContents_1",_STRINGS,ids))

	# Deprecated
	get_contents = getContents

	# Deprecated
	def getContent( self,id):
		return self.getContents([id])[id]

	# Deprecated
	get_content = getContent

	def getValues(self,ids):
		return _unsplit(ids,self._dom.call("GetValues_1",_STRINGS,ids))

	get_values = getValues

	def getValue( self,id):
		return self.get_values([id])[id]

	get_value =  getValue

	def getMarks(self,ids):
		return _unsplit(ids,self._dom.call("GetMarks_1",_STRINGS,ids))

	get_marks = getMarks

	def getMark( self,id):
		return self.get_marks([id])[id]

	get_mark = getMark

	# Deprecated
	def setContents(self,ids_and_contents):
		[ids,contents] = _split(ids_and_contents)

		self._dom.call("SetContents_1",_VOID,ids,contents)

	# Deprecated
	set_contents = setContents

	# Deprecated
	def setContent(self,id,content):
		self.set_contents({id: content})

	# Deprecated
	set_content = setContent

	def setValues(self,ids_and_values):
		[ids,values] = _split(ids_and_values)

		self._dom.call("SetValues_1",_VOID,ids,values)

	set_values = setValues

	def setValue(self,id,value):
		self.set_values({id: value})

	set_value = setValue

	def setMarks(self,ids_and_marks):
		[ids,marks] = _split(ids_and_marks)

		self._dom.call("SetMarks_1",_VOID,ids,marks)

	set_marks = setMarks

	def setMark(self,id,mark):
		self.set_marks({id: mark})

	set_mark = setMark

	def _handleClasses(self,variant,idsAndClasses):
		[ids,classes] = _split(idsAndClasses)

		self._dom.call("HandleClasses_1",_VOID,variant,ids,classes)

	def addClasses(self,ids_and_classes):
		self._handleClasses("Add",ids_and_classes)

	add_classes = addClasses

	def removeClasses(self,ids_and_classes):
		self._handleClasses("Remove",ids_and_classes)

	remove_classes = removeClasses		

	def toggleClasses(self,ids_and_classes):
		self._handleClasses("Toggle",ids_and_classes)

	toggle_classes = toggleClasses

	def addClass(self,id,clas ):
		self.addClasses({id: clas})

	add_class = addClass

	def removeClass(self,id,class_ ):
		self.removeClasses({id: class_})

	remove_class	= removeClass

	def toggleClass(self,id,clas ):
		self.toggleClasses({id: clas})

	toggle_class = toggleClass

	def enableElements(self,ids):
		self._dom.call("EnableElements_1",_VOID,ids )

	enable_elements = enableElements		

	def enableElement(self,id):
		self.enableElements([id] )

	enable_element = enableElement		

	def disableElements(self,ids):
		self._dom.call("DisableElements_1",_VOID,ids )

	disable_elements = disableElements		

	def disableElement(self,id):
		self.disableElements([id])

	disable_element = disableElement

	def setAttribute(self,id,name,value ):
		self._dom.call("SetAttribute_1",_VOID,id,name,str(value) )

	set_attribute = setAttribute		

	def getAttribute(self,id,name):
		return self._dom.call("GetAttribute_1",_STRING,id,name )

	get_attribute = getAttribute		

	def removeAttribute(self,id,name ):
		self._dom.call("RemoveAttribute_1",_VOID,id,name )

	remove_attribute = removeAttribute

	def setProperty(self,id,name,value ):
		self._dom.call("SetProperty_1",_VOID,id,name,value )

	set_property = setProperty		

	def getProperty(self,id,name ):
		return self._dom.call("GetProperty_1",_STRING,id,name )

	get_property = getProperty		

	def focus(self,id):
		self._dom.call("Focus_1",_VOID,id)

	def parent(self,id):
		return self._dom.call("Parent_1",_STRING,id)

	def firstChild(self,id):
		return self._dom.call("FirstChild_1",_STRING,id)

	first_child = firstChild

	def lastChild(self,id):
		return self._dom.call("LastChild_1",_STRING,id)

	last_child = lastChild

	def previousSibling(self,id):
		return self._dom.call("PreviousSibling_1",_STRING,id)

	previous_sibling = previousSibling		

	def nextSibling(self,id):
		return self._dom.call("NextSibling_1",_STRING,id)

	next_sibling = nextSibling

	def scrollTo(self,id):
		self._dom.call("ScrollTo_1",_VOID,id)

	scroll_to = scrollTo

	def debugLog(self,switch=True):
		self._dom.call("DebugLog_1",_VOID,"true" if switch else "false")

	def log(self,message):
		self._dom.call("Log_1",_VOID,message)



""" 
MIT License

Copyright (c) 2018 Claude SIMON (https://q37.info/s/rmnmqd49)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import XDHq, XDHqSHRD
from threading import Thread, Lock
import inspect, time, socket, signal, sys, os

from XDHq import set_supplier, get_app_url, l

setSupplier = set_supplier
getAppURL = get_app_url

# Options entries
O_CALLBACKS_PREFIX_ = "CallbacksPrefix"
O_CONNECT_ACTION_AFFIX_ = "ConnectActionName"
O_PREPROCESS_ACTION_NAME_ = "PreprocessActionName"
O_POSTPROCESS_ACTION_NAME_ = "PostprocessActionName"

DEFAULT_CALLBACK_PREFIX = "atk"

options_ = {
  # Prefix of the name of a callback corresponding to an action
  # if absent from callbacks dict.
  # If 'None' or '', callbacks for the actions must be in callbacks dict.
  O_CALLBACKS_PREFIX_: DEFAULT_CALLBACK_PREFIX,
  # The affix for the connect action.
  O_CONNECT_ACTION_AFFIX_: "",
  # NOTA: there is no default callback name for below 2 actions.
  # They must be specifically defined in the callbacks dict.
  # Name of the preprocessing action.
  O_PREPROCESS_ACTION_NAME_: "_Preprocess",
  # Name of the postprocessing action.
  O_POSTPROCESS_ACTION_NAME_: "_Postprocess"
}

if sys.version_info[0] == 2:
	import __builtin__ as builtins
else:
	import builtins

# Overriding some functions for the Dev environment.
if XDHqSHRD.isDev():
	if "openpyxl" in sys.modules :
		defaultXLFunction = sys.modules['openpyxl'].load_workbook
		sys.modules['openpyxl'].load_workbook = lambda filename, **kwargs: defaultXLFunction(XDHq.get_asset_filename(filename), **kwargs)
	
	defaultBuiltinsFunction = builtins.open
	builtins.open = lambda filename, *args, **kwargs: defaultBuiltinsFunction(XDHq.get_asset_filename(filename), *args, **kwargs)

def signal_handler(sig, frame):
  sys.exit(0)

signal.signal(signal.SIGINT, signal_handler)

def createXML(root_tag):
	return XDHq.XML(root_tag)

create_XML = createXML	


def createHTML(root_tag=""):	# If 'root_tag' is empty, there will be no root tag in the tree.
	return XDHq.XML(root_tag)

create_HTML = createHTML

broadcastAction = XDHq.broadcastAction
broadcast_action = broadcastAction


def _call(func, userObject, dom, id, action):
	amount = len(inspect.getfullargspec(func).args)
	args = []

	if not(userObject):
		amount += 1

	if amount == 4:
		args.insert(0, action)

	if amount >= 3:
		args.insert(0, id)

	if amount >= 2:
		args.insert(0 ,dom)

	if userObject and ( amount >= 1 ):
		args.insert(0, userObject)

	return func(*args)

def _is_jupyter():
	try:
			shell = get_ipython().__class__.__name__
			if shell == 'ZMQInteractiveShell':
					return True   # Jupyter notebook or qtconsole
			elif shell == 'TerminalInteractiveShell':
					return False  # Terminal running IPython
			else:
					return 'google.colab' in str(get_ipython())  # Other type (?)
	except NameError:
			return False      # Probably standard Python interpreter

def worker(userCallback, dom, callbacks, callingGlobals):
	args=[]
	userObject = None

	if ( userCallback != None ):
		if ( not(inspect.isclass(userCallback)) and len(inspect.getfullargspec(userCallback).args) == 1 ):
			args.append(dom)

		userObject = userCallback(*args)

	while True:
		[action,id] = dom.getAction()

		if action == "":
			if _is_jupyter():
				dom.disableElement("XDHStyle")
			else:
				dom.disableElement("XDHStyleJupyter")

			if XDHqSHRD.isDev():
				dom.debugLog(True)

		if action == ""\
			or not options_[O_PREPROCESS_ACTION_NAME_] in callbacks\
			or _call(callbacks[options_[O_PREPROCESS_ACTION_NAME_]], userObject, dom, id, action) in [None, True]:

			callback = None

			if action in callbacks:
				callback = callbacks[action]
			elif options_[O_CALLBACKS_PREFIX_]:
				callbackName = options_[O_CALLBACKS_PREFIX_] + ( options_[O_CONNECT_ACTION_AFFIX_] if action == "" else action )

				if callbackName in callingGlobals:
					callback = callingGlobals[callbackName]
			
			if callback:
				dom.callbackReturnValue = _call(callback, userObject, dom, id, action )
				
				if options_[O_POSTPROCESS_ACTION_NAME_] in callbacks:
					_call(callbacks[options_[O_POSTPROCESS_ACTION_NAME_]],userObject, dom, id, action)
			else:
				dom.alert("\tDEV ERROR: missing callback for '" + action + "' action!") 
	
	# l() # Exiting thread, closing corresponding instance.

def _callback(userCallback, callbacks, callingGlobals, instance):
	thread = Thread(target=worker, args=(userCallback, XDHq.DOM(instance), callbacks, callingGlobals))
	thread.daemon = True
	thread.start()
	return thread

def _jupyter_supplier(url):
	global _url, _intraLock

	_url = url.replace('http://', 'https://')
	_intraLock.release()

if _is_jupyter():
	import IPython
	global _intraLock, _thread, _jupyterWidth, _jupyterHeight

	_jupyterWidth = "100%"
	_jupyterHeight = "200px"

	_intraLock = Lock()
	XDHq.set_supplier(_jupyter_supplier)

	_thread = None

	def setJupyterWidth(width):
		global _jupyterWidth

		_jupyterWidth = width

	def setJupyterHeight(height):
		global _jupyterHeight

		_jupyterHeight = height

	def terminate():
		global _thread

		if _thread != None:
			XDHq.setBye(True)
			_thread.join()
			XDHq.setBye(False)
			_thread = None


def _launch(callbacks, callingGlobals, userCallback, headContent):
	if callbacks == None:
		callbacks = {}

	if callingGlobals == None:
		callingGlobals = {}

	try:
		XDHq.launch(_callback, userCallback, callbacks, callingGlobals, headContent)
	except socket.timeout:
		pass


def retrieve_(var, id, globals):
  if ( var == None ) and ( id in globals ):
    return globals[id]
  
  return var	
	

def launch(callbacks = None, *, globals = None,  userCallback = None, headContent = None):
	if globals != None:
		callbacks = retrieve_(callbacks, "ATK_CALLBACKS", globals)
		userCallback = retrieve_(userCallback, "ATK_USER", globals)
		headContent = retrieve_(headContent, "ATK_HEAD", globals)

	if _is_jupyter():
		global _intraLock, _thread

		terminate()
		
		_intraLock.acquire()
		_thread = Thread(target=_launch, args=(callbacks, globals, userCallback, headContent))
		_thread.daemon = True
		_thread.start()

		_intraLock.acquire()		

		iframe = IPython.display.IFrame(src = _url, width = _jupyterWidth, height = _jupyterHeight)
		_intraLock.release()
		return iframe
	else:
		_launch(callbacks, globals, userCallback, headContent)

def options(options = None):
  if options != None:
    global options_

    for key in options:
      if not key in options_:
        raise Exception(f"Unknown option '{key}'!")
      else:
        options_[key] = options[key]

  return options_




"""
MIT License

Copyright (c) 2018 Claude SIMON (https://q37.info/s/rmnmqd49)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

_VERSION = "0.13.4"

import XDHqSHRD
from XDHqSHRD import getEnv

import inspect, os, socket, sys, threading, time

if sys.version_info[0] == 2:
	import XDHqFaaS2
	l = XDHqFaaS2.l
	_writeUInt = XDHqFaaS2.writeUInt
	_writeString = XDHqFaaS2.writeString
	_readUInt = XDHqFaaS2.readUInt
	_getString = XDHqFaaS2.getString
	_exitThread = XDHqFaaS2.exitThread
else:
	import XDHqFaaS3
	l = XDHqFaaS3.l
	_writeUInt = XDHqFaaS3.writeUInt
	_writeString = XDHqFaaS3.writeString
	_readUInt = XDHqFaaS3.readUInt
	_getString = XDHqFaaS3.getString
	_exitThread = XDHqFaaS3.exitThread

_bye = False	# For use in Jupiter notebooks, to quit an application.

_DEFAULT_SUPPLIER_LABEL = "auto"

class _Supplier:
	current = None

	_actions = {
		"none": lambda url : None,
		_DEFAULT_SUPPLIER_LABEL: XDHqSHRD.open,
		"qrcode": lambda url: XDHqSHRD.open( '"' + url + '&_supplier=qrcode"'),
		"jupyter": lambda url : None
	}

def _supply(url):
	supplier = getEnv("ATK").strip().lower() or _Supplier.current or _DEFAULT_SUPPLIER_LABEL

	while True:
		supplier = _Supplier._actions[supplier](url) if isinstance(supplier, str) else supplier(url)

		if not supplier:
			break;
			
def set_supplier(supplier = None):
	_Supplier.current = supplier

_FAAS_PROTOCOL_LABEL = "4c837d30-2eb5-41af-9b3d-6c8bf01d8dbf"
_FAAS_PROTOCOL_VERSION = "1"
_MAIN_PROTOCOL_LABEL = "22bb5d73-924f-473f-a68a-14f41d8bfa83"
_MAIN_PROTOCOL_VERSION = "0"
_SCRIPTS_VERSION = "0"

_FORBIDDEN_ID = -1
_CREATION_ID = -2
_CLOSING_ID = -3
_HEAD_RETRIEVING_ID = -4

_BROADCAST_ACTION_ID = -3
_HEAD_SENDING_ID = -4

_writeLock = threading.Lock()

_readLock = threading.Lock()	# Global read lock.
_readLock.acquire()


def _waitForInstance():
	_readLock.acquire()


def _instanceDataRead():
	_readLock.release()


_url = ""

class _Instance:
	def __init__(self,thread_retriever,id):
		# https://github.com/epeios-q37/atlas-python/pull/7 (Condition -> Lock)
		self._readLock = threading.Lock()	#Per instance read lock.
		self._readLock.acquire()
		self.quit = False
		self.id = id
		self.thread = thread_retriever(self)
		self.language = None
	def getId(self):
		return self.id
	def waitForData(self):
		self._readLock.acquire()
		if self.quit:
			_instanceDataRead()
			_exitThread()
	def dataAvailable(self):
		self._readLock.release()

def isTokenEmpty():
	return not _token or _token[0] == "&"

def writeUInt(value):
	global _socket
	_writeUInt( _socket, value, lambda: _bye )

def writeSInt(value):
	writeUInt( ( ( -value - 1 ) << 1 ) | 1 if value < 0 else value << 1 )

def writeString(string):
	global _socket
	_writeString(_socket, string, lambda: _bye)

def writeStrings(strings):
	writeUInt(len(strings))

	for string in strings:
		writeString(string)

def readUInt():
	global _socket
	return _readUInt( _socket, lambda: _bye)

def readSInt():
	value = readUInt()
	return -( ( value >> 1 ) + 1 ) if ( value & 1 ) else value >> 1

def getString():
	global _socket
	return _getString(_socket, lambda: _bye)

def getStrings():
	amount = readUInt()
	strings = []

	while amount:
		strings.append(getString())
		amount -= 1

	return strings

def _dismiss(id):
	with _writeLock:
		writeSInt(id)
		writeString("#Dismiss_1")

def _report(message):
	with _writeLock:
		writeSInt(-1)
		writeString("#Inform_1")
		writeString(message)

def _init():
	global _token, _socket, _wAddr, _wPort, _cgi
	pAddr = "faas.q37.info"
	pPort = 53700
	_token = ""
	_wAddr = ""
	_wPort = ""
	_cgi = "xdh"

	pAddr = getEnv("ATK_PADDR", pAddr)
	pPort = int(getEnv("ATK_PPORT", str(pPort)))
	_wAddr = getEnv("ATK_WADDR", _wAddr)
	_wPort = getEnv("ATK_WPORT", _wPort)

	if _wAddr == "":
		_wAddr = pAddr

	if _wPort != "":
		_wPort = ":" + _wPort

	if isTokenEmpty():
		_token = getEnv("ATK_TOKEN")

	if _token:
		_token = "&" + _token

	_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	
	print("Connection to '" + str(pAddr) + ":" + str(pPort) + "'...")

	try:
		_socket.connect((pAddr,pPort))
	except:
		sys.exit("Unable to connect to '" + str(pAddr) + ":" + str(pPort) + "'!")
	else:
		print("Connected to '" + str(pAddr) + ":" + str(pPort) + "'.")

	_socket.settimeout(1)	# In order to quit an application, in Jupyter notebooks.		

def _handshakeFaaS():
	with _writeLock:
		writeString(_FAAS_PROTOCOL_LABEL)
		writeString(_FAAS_PROTOCOL_VERSION)
		writeString("PYH " + _VERSION)

	error = getString()

	if error:
		sys.exit(error)

	notification = getString()

	if notification:
		print(notification)

def _handshakeMain():
	with _writeLock:
		writeString(_MAIN_PROTOCOL_LABEL)
		writeString(_MAIN_PROTOCOL_VERSION)
		writeString(_SCRIPTS_VERSION)

	error = getString()

	if error:
		sys.exit(error)

	notification = getString();

	if notification:
		print(notification)

def _handshakes():
	_handshakeFaaS()
	_handshakeMain()

def _ignition():
	global _token, _url
	with _writeLock:
		writeString( _token)
#		writeString(_headContent)	# Dedicated call from proxy since protocol v1
		writeString(_wAddr)
		writeString("")	# Currently not used; for future use.

	_token = getString()

	if isTokenEmpty():
		sys.exit(getString())

	if ( _wPort != ":0" ):
		_url = getString()

		print(_url)
		print("".rjust(len(_url),'^'))
		print("Open above URL in a web browser (click, right click or copy/paste). Enjoy!\n")

		_supply(_url)

def _serve(callback, userCallback, callbacks, callingGlobals):
	while True:
		id = readSInt()
		
		if id == _FORBIDDEN_ID:	# Should never happen. 
			sys.exit("Received unexpected undefined command id!")
		if id == _CREATION_ID:    # Value reporting a new session.
			id = readSInt()  # The id of the new session.

			if id in _instances:
				_report("Instance of id '" + str(id) + "' exists but should not !")

			_instances[id] = _Instance(lambda instance : callback(userCallback, callbacks, callingGlobals, instance), id)
		elif id == _CLOSING_ID:	# Value instructing that a session is closed.
			id = readSInt()

			if not id in _instances:
				_report("Instance of id '" + str(id) + "' not available for destruction!")
			else:
				instance = _instances.pop(id)
				instance.quit = True
				instance.dataAvailable()
				_waitForInstance()
				instance = None # Without this, instance will only be destroyed
												# when 'instance" is set to a new instance.
		elif id == _HEAD_RETRIEVING_ID:
			writeSInt(_HEAD_SENDING_ID)
			writeString(_headContent)
		elif not id in _instances:
			_report("Unknown instance of id '" + str(id) + "'!")
			_dismiss(id)
		else:
			instance = _instances[id]

			if instance.language is None:
				instance.language = getString()
			else:
				instance.dataAvailable()
				_waitForInstance()


def launch(callback, userCallback, callbacks, callingGlobals, headContent):
	global _headContent, _instances

	if headContent is None:
		if not "_headContent" in globals():
			_headContent = "" 
	else:
		_headContent = headContent
	
	_instances = {}

	_init()

	_handshakes()

	_ignition()

	_serve(callback, userCallback, callbacks, callingGlobals)

def get_app_url(id=""):
	return _url + ("&_id=" + str(id) if id else "") 

def broadcastAction(action, id = ""):
	with _writeLock:
		writeSInt(_BROADCAST_ACTION_ID)
		writeString(action)
		writeString(id)

class DOM_FaaS:
	_firstLaunch = True

	def __init__(self, instance):
		self.instance = instance

	def _waitForData(self):
		self.instance.waitForData()
		
	def _standBy(self):
		with _writeLock:
			writeSInt(self.instance.getId())
			writeString("#StandBy_1")

	def getAction(self):
		if self._firstLaunch:
			self._firstLaunch = False
		else:
			self._standBy()

		self._waitForData()

		[id,action] = [getString(),getString()]

		_instanceDataRead()

		return [action,id]

	def call(self, command, type, *args):
		if self.instance.quit:
			"""
			Below function unlocks the main thread,
			and exits the thread corresponding
			to the current instance.
			"""
			self._waitForData()

		with _writeLock:
			writeSInt(self.instance.getId())
			writeString(command)

			writeUInt(type)

			for arg in args:
				if isinstance(arg,str):
					writeUInt(XDHqSHRD.RT_STRING)
					writeString(arg)
				else:
					writeUInt(XDHqSHRD.RT_STRINGS)
					writeStrings(arg)

			writeUInt(XDHqSHRD.RT_VOID)	# To report end of argument list.

		if type == XDHqSHRD.RT_STRING:
			self._waitForData()
			string = getString()
			_instanceDataRead()
			return string
		elif type == XDHqSHRD.RT_STRINGS:
			self._waitForData()
			strings = getStrings()
			_instanceDataRead()
			return strings
		elif type != XDHqSHRD.RT_VOID:
			sys.exit("Unknown return type !!!")

def setBye(value):
	global _bye

	_bye = value
